#---------------------------------------------------------------------------+
# Licensed Materials - Property of IBM
# Rational ClearCase
# Copyright IBM Corp. 2005, 2006.  All Rights Reserved.
# US Government Users Restricted Rights -
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract
# with IBM Corp.
#---------------------------------------------------------------------------+

package ClearCase::RD_utils;

use 5.8.6;
use Exporter;
use strict;
use URI;
use Encode;
use Config;
use XML::Parser;

our @ISA = qw(Exporter);
our @EXPORT = ('set_script_name',
               'set_binmode',
               '$cleartool', 
               '$errno',
               'pretty_print',
               'remove_external_whitespace',
               'do_cmd',
               'native_cp_to_internal_utf8',
               'internal_utf8_to_native_cp',
               'get_cwv_tag',
               'cat_view_dir_and_pname',
               'chase_symlinks',
               'convert_pname_to_uri',
               'convert_uri_to_pname',
               'dynamic_view_is_started',
               'get_view_info_from_pname',
               'get_object_meta_type',
               'get_baselines_from_du_file',
               'get_oref_info_from_uri',
               
               # File::Spec::Functions wrappers
               'canonpath',
               'catdir',
               'catfile',
               'catpath',
               'file_name_is_absolute',
               'rel2abs',
               'rootdir',
               'splitdir',
               'splitpath',
               
               # Cwd wrappers
               'RD_getcwd',
               'RD_chdir',
               
               # Unary file test operator wrappers
               'file_test');

# Name of the cleartool executable.  User must either (a) ensure that
# cleartool is in the path or (b) specify an absolute path in $cleartool.
our $cleartool = 'cleartool';

# C-style errno variable and values; intended for use only by CQ hook code to
# render interesting messages.  Interactively run scripts (such as du_tool.pl)
# should be content with internally-emitted output.
#
# Values are defined as variables rather than constants to allow hook code to
# 'require' this module and thus use it only when needed.  They are
# deliberately NOT exported; access as 
# "if ($errno == $ClearCase::RD_utils::ERRNO_INT_ERR)".
our $ERRNO_INT_ERR = 0;          # Generic internal error
our $ERRNO_ACCESS = 1;           # Can't access file/directory
our $ERRNO_NOT_IN_VIEW = 2;      # Not in a ClearCase view
our $ERRNO_LOOP = 2;             # Circular symlinks
our $ERRNO_IS_DIRECTORY = 3;     # Pathname is a directory
our $ERRNO_VIEW_PVT = 4;         # Pathname is a view-private object
our $ERRNO_NOT_VERSION_FILE = 5; # Pathname is not a CC file version
our $ERRNO_CHECKOUT = 6;         # File is checked out

our $errno = $ERRNO_INT_ERR;

# Message catalog; to internationalize this script, translate the messages,
# but leave the names intact.
{
    package ClearCase::RD_utils::msgs;

    # Message formatting strings
    use constant WARNING_PREFIX => "Warning: ";
    use constant ERROR_PREFIX => "Error: ";
    
    # General-purpose messages
    use constant BAD_CALL => "Invalid arguments to \"%s\".\n";
    use constant CANT_EXECUTE => "Cannot execute \"%s\": %s\n";
    use constant UNEXPECTED_OUTPUT => "Unexpected output from command \"%s\"\n";
    use constant TOO_MANY_LINKS => 
        "Cannot resolve all symbolic links from \"%s\"\n";
    use constant NO_USER => 
        "Unable to determine the current user's identity.\n";
    use constant CANT_CONVERT_FROM_TO =>
        "Unable to convert \"%s\" from \"%s\" to \"%s\": %s\n";
    use constant CANT_CONVERT_TO_UTF8 =>
        "Unable to convert \"%s\" from \"%s\" to Perl's " .
        "internal UTF-8 encoding: %s\n";
    use constant CANT_CONVERT_FROM_UTF8 =>
        "Unable to convert \"%s\" from Perl's internal UTF-8 encoding to" .
        "\"%s\": %s\n";

    # Command line syntax 
    use constant USAGE_PREFIX => "Usage:\n";
    use constant USAGE_SUFFIX => 
        "For a detailed explanation of usage options, see the " .
        "du_tool reference page in the Rational ClearCase " .
        "Command Reference Manual.\n\n";
    use constant INVALID_OPTION => "Invalid option: %s\n";
    use constant CANT_GET_ARGS => "Cannot read arguments.\n";                 
    use constant ONE_OPERATION => "Must specify exactly one operation.\n";
    use constant GENERATE_REQD_FLAGS =>
        "--directory, --name and --version are requred when using " .
        "--generate.\n";
    use constant GENERATE_ONLY_FLAGS => 
        "--directory, --name and --version are allowed only with --generate.\n";
    use constant NEED_VIEW_CTX => "This tool must be run in a view context.\n";
    use constant NOT_SNAPSHOT_VIEW => "\"%s\" is not a snapshot view.\n";
    use constant NOT_DYNAMIC_VIEW => "\"%s\" is not a dynamic view.\n";
    use constant DECORATE_ONLY_FLAGS => 
        "--instance, --baseline, --logvendor and --buildlog are valid only " .
        "with --decorate.\n";
    use constant VENDOR_REQUIRES_BUILDLOG =>
        "Cannot specify --logvendor without also specifying --buildlog.\n";
    use constant VIEW_WITH_VERIFY => "May not specify a view with --verify.\n";
    use constant ONE_VIEW => "May specify only one view.\n";
    use constant OUT_WITH_VERIFY => 
        "May not specify an output file with --verify.\n";
    use constant NO_MULTIPLE_VIEW_PATHS => 
        "Path \"%s\" is not within the \"%s\" view.\n";
    use constant NO_RELATIVE_PATHS =>
        "Relative path \"%s\" not allowed " .
        "when invoked with --tag or --pname.\n";

    # Current working directory stuff
    use constant DIR_ACCESS => "Cannot access directory \"%s\": %s\n";

    # File stuff
    use constant CANT_CLOSE => "Cannot close file \"%s\": %s\n";
    use constant CANT_OPEN => "Cannot open file \"%s\": %s\n";                 
    use constant CANT_UNLINK => "Cannot remove output file \"%s\": %s\n";
    use constant FILE_EXISTS => "The file \"%s\" already exists.\n";
    use constant PNAME_DOESNT_EXIST => "\"%s\" does not exist.\n";
    use constant DIR_NOT_ALLOWED => "Cannot operate on directory \"%s\".\n";
    use constant VIEW_PVT_NOT_ALLOWED => 
        "Cannot operate on view private file \"%s\".\n";
    use constant NON_VERSION_NOT_ALLOWED => 
        "Cannot operate on \"%s\" because it is not a versioned object.\n";
    use constant NO_CHECKOUTS => 
        "Cannot operate on checked-out version \"%s\".\n";
    use constant CANT_CHASE_LINK => 
        "Cannot chase apparent symbolic link \"%s\": %s\n";
    use constant CANT_CONVERT_PNAME =>
        "Cannot decorate file \"%s\" at line %d.\n";
    use constant VALID_FILE => "Input XML is valid.\n";
    use constant BAD_FILE => "Parsing failed: %s\n";
    use constant REMOVING_BAD_OUTPUT => 
        "Removing empty/incomplete output file \"%s\".\n";
    use constant PATH_NOT_IN_VIEW =>
        "The pathname \"%s\" is not within a view.\n";

    # XML stuff
    use constant BAD_VERSION => "Unsupported XML version: \"%s\"\n";
    use constant BAD_ENCODING => "Unsupported XML encoding \"%s\"\n";
    use constant MISSING_ATTR =>
                  "Missing attribute \"%s\" (element \"%s\", line %d).\n";
    use constant NO_FILE_ELEMENTS => 
        "Input does not contain any file entries on which to operate.\n";

    # URI stuff
    use constant BAD_URI_SCHEME => "Invalid URI scheme: \"%s\"\n";
    use constant MISSING_URI_QUERY_NAME => 
        "Missing URI %s query name/value pair.\n";
    use constant MALFORMED_URI_OREF => 
        "The URI \"%s\" does not contain a valid OREF.\n";

    # Initial generation stuff
    use constant FOUND_NO_FILES =>
        "Found no acceptable files for which to generate output.\n";
    use constant NOT_A_DIRECTORY =>
        "The path \"%s\" is not a directory in the given view.\n";
    use constant SKIPPING_VIEW_PVT_OBJ =>
        "Skipping view private object \"%s\".\n";
    use constant SKIPPING_NULL_META_TYPE =>
        "Skipping \"%s\" because it appears not to be a ClearCase object.\n";
    use constant SKIPPING_NON_VERSION_TYPE =>
        "Skipping \"%s\" because it appears not to be a versioned object.\n";

    # Known issues
    use constant DIR_HAS_TRAILING_5C_BYTE => 
        "The final character in the directory \"%s\" has a trailing " . 
        "0x5c byte.  This tool cannot safely handle such directory names " .
        "on Windows.\n";

} # Messages sub-package

# Enumerated inputs for file_test
use constant FILE_TEST_DIRECTORY => 0; # -d
use constant FILE_TEST_EXISTS => 1;    # -e
use constant FILE_TEST_FILE => 2;      # -f
use constant FILE_TEST_LINK => 3;      # -l

# Useful info that is required across multiple XML parsing activities
# or that would be computationally wasteful to evaluate when processing an XML
# file containing more than a single file element. (Example: access as
# ClearCase::RD_utils::XML_DOC_ELEMENT)
use constant XML_DOC_ELEMENT => 'tracking:deploymentUnit';
use constant XML_APPNAME_ATTR => 'name';
use constant XML_APPVER_ATTR => 'version';
use constant XML_XSI_ATTR => 'xmlns:xsi';
use constant XML_XSI_VAL => 'http://www.w3.org/2001/XMLSchema-instance';
use constant XML_TRACKING_ATTR => 'xmlns:tracking';
use constant XML_TRACKING_VAL => 
    'http://www.ibm.com/deployment/schema/deploymentTracking.xsd';
use constant XML_INSTANCE_ATTR => 'instanceID';
use constant XML_FILE_ELEMENT => 'file';
use constant FILE_PNAME_ATTR => 'path';
use constant FILE_ORIG_ATTR => 'origPath';
use constant FILE_OREF_ATTR => 'versionURI';
use constant FILE_CSRULE_ATTR => 'csrule';
use constant FILE_XNAME_ATTR => 'xname';
use constant XML_ORIGIN_ELEMENT => 'CCOrigin';
use constant XML_CS_ELEMENT => 'configSpec';
use constant XML_BASELINE_ELEMENT => 'baseline';
use constant BASELINE_NAME_ATTR => 'baselineName';
use constant BASELINE_AUTH_ATTR => 'originAuthority';
use constant BASELINE_AUTH_VAL => 'UserProvided';
use constant XML_HISTORY_ELEMENT => 'history';
use constant HISTORY_OPERATION_ATTR => 'changeOperation';
use constant HISTORY_OP_GENERATE => 'Generate';
use constant HISTORY_OP_DECORATE => 'Decorate';
use constant HISTORY_OP_RESOLVE => 'Resolve';
use constant HISTORY_NOTE_ATTR => 'changeNote';
use constant HISTORY_HOST_ATTR => 'hostname';
use constant HISTORY_USER_ATTR => 'user';
use constant HISTORY_TS_ATTR => 'datetime';
use constant XML_BLDORGN_ELEMENT => 'buildOrigin';
use constant XML_BLDLG_ELEM => 'buildLog';
use constant BLDLG_VENDOR_ATTR => 'vendor';
use constant EMPTY_XML_ELEMENTS => {
    XML_FILE_ELEMENT() => 1,
    XML_BASELINE_ELEMENT() => 1,
    XML_HISTORY_ELEMENT() => 1,
    XML_BLDLG_ELEM() => 1,
};
# This hash contains elements just inside the document level that are newly
# emitted during a decorate operation.  If encountered during decoration, they
# are elided in favor of new elements.
use constant DECORATE_CREATED_ELEMENTS => {
    XML_ORIGIN_ELEMENT() => 1,
    XML_BLDORGN_ELEMENT() => 1,
};

my $this_script = undef;

################################################################################
# set_script_name -       Set the name of the script to be displayed in
#                         error/warning messages emitted by pretty_print()
#                         and any other function in this package.
#
# IN   new_name           The new script name.
sub set_script_name($) 
{
    my ($new_name) = @_;

    $this_script = $new_name;
} # set_script_name


# Currently set encodings for each of the system-definied input/output
# filehandles.  If not defined, handle is set to the codepage returned by
# get_codepage().
my $stdin_encoding = undef;
my $stdout_encoding = undef;
my $stderr_encoding = undef;


# Enumerated inputs for set_binmode
use constant RD_STDIN => 0;
use constant RD_STDOUT => 1;
use constant RD_STDERR => 2;


################################################################################
# set_binmode -           Set the binmode for STDIN/OUT/ERR.
#                         You MUST use this wrapper around binmode() if you
#                         wish to do piped input/output in something other
#                         than the system's native encoding (e.g. UTF-8 XML)
#                         and preserve good behavior in pretty_print().
#
# IN   $fh                One of [RD_STDIN | RD_STDOUT | RD_STDERR].
# IN   $mode              Mode string to pass to Perl's binmode() function.
#                         Use an empty string to use the system's native
#                         encoding.
sub set_binmode($$) 
{
    my ($fh, $mode) = @_;
    my $actual_fh = undef;
    my $encoding_ref = undef;

    if ($fh == RD_STDIN) {
        $actual_fh = \*STDIN;
        $encoding_ref = \$stdin_encoding;
    }
    elsif ($fh == RD_STDOUT) {
        $actual_fh = \*STDOUT;
        $encoding_ref = \$stdout_encoding;
    }
    elsif ($fh == RD_STDERR) {
        $actual_fh = \*STDERR;
        $encoding_ref = \$stderr_encoding;
    }
    else {
        die;
    }

    my $actual_mode = $mode;

    if ($actual_mode) {
        $$encoding_ref = $actual_mode;
    }
    else {
        undef($$encoding_ref) if ($encoding_ref);
        $actual_mode = ':encoding(' . get_codepage() . ')';
    }
    
    binmode($$actual_fh, $actual_mode);

} # set_bimode


# Enumerated inputs for pretty_print.
use constant MSG_OK => 0;
use constant MSG_WARN => 1; 
use constant MSG_ERR => 2;


################################################################################
# pretty_print            Print consistently-formatted output to
#                         STDOUT/STDERR.
#
# IN   stream             Target output stream.  Input MUST be one of the
#                         ClearCase::RD_utils::MSG_* constants.  (If 
#                         unspecified, defaults to MSG_OK.)
#
#                         Output format for each input:
#                           MSG_OK:   "$msg" to STDOUT
#                           MSG_WARN: "$this_script: Warning: $msg" to STDERR
#                           MSG_ERR:  "$this_script: Error: $msg" to STDERR
# IN   msg                Message to print.
# IN   varargs            Any printf-formatted string arguments for msg.
sub pretty_print($$@)
{
    my ($stream, $msg, @varargs) = @_;
    
    die if (!$msg);

    my $fh = \*STDERR;
    my $fh_encoding_ref = \$stderr_encoding;
    my $prefix = ($this_script ? "$this_script: " : "");
    
    if ($stream == MSG_ERR) {
        $prefix .= ClearCase::RD_utils::msgs::ERROR_PREFIX;
    }
    elsif ($stream ==  MSG_WARN) {
        $prefix .= ClearCase::RD_utils::msgs::WARNING_PREFIX;
    }
    else{
        $prefix = "";
        $fh = \*STDOUT;
        $fh_encoding_ref = \$stdout_encoding;
    }

    # Output MUST be in native codepage
    if ($$fh_encoding_ref) {
        binmode($$fh, ':encoding(' . get_codepage() . ')');
    }

    printf { $$fh } $prefix . $msg, @varargs;

    # Revert to client-defined encoding
    if ($$fh_encoding_ref) {
        binmode($$fh, $$fh_encoding_ref);
    }
} # pretty_print


################################################################################
# remove_external_whitespace -
#                         Remove leading & trailing whitespace from a string.
#
# IO   str                String to be edited.
sub remove_external_whitespace(\$) 
{
    my ($str) = @_;

    return unless ($str && $$str);

    $$str =~ s/^\s+//;
    $$str =~ s/\s+$//;                
    
} # remove_external_whitespace


################################################################################
# do_cmd -                Perform a system command.  
#
#                         NOTE WELL: This function does NOT work on Windows if
#                         the process' current working directory ends in a
#                         multibyte character with a trailing 0x5c byte!
#
# IN   cmd                Command to execute.
# OUT  cmd_output_ref     Output generated by command.  May be empty; only
#                         caller knows context.  
# RESULT:                 0 if command fails.  Non-zero otherwise.
sub do_cmd($\$) 
{
    my ($cmd, $cmd_output_ref) = @_;

    if (!$cmd || !$cmd_output_ref) {
        pretty_print(MSG_ERR, ClearCase::RD_utils::msgs::BAD_CALL, "do_cmd");
        return 0;
    }

    # Perl on Windows cannot call out to the shell (or read a program's output
    # from a pipe created with open()) if the current working directory ends
    # with a 0x5c byte.  The errno value in $! generally complains about a bad
    # file descriptor.
    #
    # The least we can do is warn when we know that we're in such a
    # situation.  We just need to make sure not to warn if the last byte
    # really IS a '\' character, and the safest way to do that is to check the
    # UTF-8-encoded working directory for a trailing '\' as well.
    my $working_dir = undef;

    if (!os_is_unix() && 
        (($working_dir = RD_getcwd()) =~ m/\\$/) &&
        !(native_cp_to_internal_utf8($working_dir) =~ m/\\$/))
    {
        pretty_print(MSG_WARN, 
                     ClearCase::RD_utils::msgs::DIR_HAS_TRAILING_5C_BYTE, 
                     $working_dir);
    }

    $$cmd_output_ref = `$cmd`;
    chomp($$cmd_output_ref);

    if ($?) {
        # $$cmd_output_ref will contain any messages that the command dumped
        # to STDOUT; those reported to STDERR have already been printed.  Exit
        # status in $? doesn't contain any useful information for the user.
        # $! does, however.
        pretty_print(MSG_ERR, 
                     ClearCase::RD_utils::msgs::CANT_EXECUTE, 
                     $cmd,
                     $!);
        return 0;
    }

    return 1;

} # do_cmd

use constant UTF8_ENCODING => 'UTF-8';

my $registry = undef;
my $region = undef;

################################################################################
# get_registry_and_region -
#                         Get this host's ClearCase registry and region.
#                         This function is intended SOLELY for use by
#                         convert_pname_to_uri().
#
# RESULT:                 External-UTF-8-encoded array: ($registry, $region)
#                         Blurts and returns undef if errors occur.
sub get_registry_and_region()
{
    return ($registry, $region) if ($registry && $region);

    my $cmd_result = undef;
    my $cmd = "$cleartool hostinfo -long";

    # The hostinfo command generates output thusly:
    #     Client: <client name>
    #       Product: <version>
    #       Operating system: <OS>
    #       Hardware type: <platform>
    #       Registry host: <rgy host>
    #       Registry region: <region>
    #       License host: <license host>
    #
    # The "Client", "Product", etc. strings may be translated, so we can't
    # search for them.  Instead, we'll rely on the fact that the registry host
    # is the 5th line of output, and the region the 6th.

    return undef if (!do_cmd($cmd, $cmd_result));

    my @lines = split(/^/, $cmd_result);
    my @rgy_tokens = split(/\:/ , $lines[4]);
    my @region_tokens = split(/\:/ , $lines[5]);
    my $tmp_registry = $rgy_tokens[1];
    my $tmp_region = $region_tokens[1];

    $tmp_registry =~ s/\s+//;
    $tmp_region =~ s/\s+//;
    
    chomp($tmp_registry);
    chomp($tmp_region);

    if (!$tmp_registry || !$tmp_region) {
        pretty_print(MSG_ERR, 
                     ClearCase::RD_utils::msgs::UNEXPECTED_OUTPUT, 
                     $cmd);
        return undef;
    }

    # Want to expand registry host to a fully-qualified name, if possible.
    my ($rgy_name, $rgy_aliases) = gethostbyname($tmp_registry);

    # Take the first alias, whatever it may be.
    ($tmp_registry) =  split(/\s+/, $rgy_aliases) if ($rgy_aliases);

    # Convert from the native CP to the external UTF-8 representation, as
    # required by convert_pname_to_uri(), which should be this function's ONLY
    # caller.
    foreach my $str ($tmp_registry, $tmp_region) {
        eval { Encode::from_to($str, 
                               get_codepage(),
                               UTF8_ENCODING, 
                               Encode::FB_CROAK) };
        next unless ($@);
        pretty_print(MSG_ERR,
                     ClearCase::RD_utils::msgs::CANT_CONVERT_FROM_TO,
                     $str,
                     get_codepage(),
                     UTF8_ENCODING,
                     $@);
        return undef;
    }

    $registry = $tmp_registry;
    $region = $tmp_region;

    return ($registry, $region);

} # get_registry_and_region


################################################################################
# get_cwv_tag -           Get the view tag of the current working view.  
#
# RESULT:                 View's tag if in a view context.  (Caller must
#                         determine view's type.) False if no view context.
#                         Blurts and returns undef if errors occur.
sub get_cwv_tag()
{
    # cleartool pwv does all of the work that we need, including
    # selecting the working directory view over the set view if the two differ.
    # It returns a translated string with the '*' character ONLY when not in a
    # view context.
    my $view_tag = undef;

    return undef unless do_cmd("$cleartool pwv -s", $view_tag);
    return ($view_tag =~ /^\*/) ? "" : $view_tag;
} # get_cwv_tag


################################################################################
# dynamic_view_is_started -
#                         Determine whether a suspected dynamic view is
#                         started.
#
# IN    tag               View tag.
# RESULT:                 Scalar context: boolean indicating whether view is
#                                         started.
#                         List context: ($view_is_started, $view_dir)
#                         Blurts and returns undef if errors occur.
sub dynamic_view_is_started($) 
{
    my ($tag) = @_;

    if (!$tag) {
        pretty_print(MSG_ERR, 
                     ClearCase::RD_utils::msgs::BAD_CALL, 
                     "dynamic_view_is_started");
        return undef;
    }

    my $viewroot = undef;

    return undef if (!do_cmd("$cleartool pwv -mvfsroot", $viewroot));

    # If MVFS is not installed, $viewroot will be empty.
    my $view_dir = $viewroot ? canonpath(catdir($viewroot, $tag)) : undef;
    my $started = 
        $view_dir ? defined(file_test(FILE_TEST_DIRECTORY, $view_dir)) : 0;
    
    return wantarray() ? ($started, $view_dir) : $started;

} # dynamic_view_is_started


################################################################################
# get_view_info_from_pname - 
#                         Get assorted information about a view from a
#                         pathname within that view.
#
# IN    pname             Absolute pathname within a view
#                         (eg. /view/my_view/src_vob/foo.cxx or any path
#                         within a set view))  This pname MUST exist, but may
#                         be any valid filesystem object (file, directory,
#                         link, etc.)
#
# RESULT:                 ($view_tag, $view_dir, $pname_in_view)
#                         
#                         $view_tag is the view's tag.
#
#                         $view_dir is the view's directory. If a dynamic
#                         view, this will always be a valid view directory
#                         (/view/my_view or M:\my_view), even if pname is
#                         within a set view.  (Snapshot view will always be
#                         returned as its directory.)
#
#                         $pname_in_view is the viewroot-relative path to the
#                         right of the view context in the input pname
#                         (/src_vob/foo.cxx in the example above).  May be
#                         empty if pname is a view directory.
#
#                         Blurts, returns undef and sets $errno if errors
#                         occur. 
sub get_view_info_from_pname($)
{
    # Default error case
    $errno = $ERRNO_INT_ERR;

    my ($pname) = @_;

    unless ($pname && file_name_is_absolute($pname)) {
        pretty_print(MSG_ERR, 
                     ClearCase::RD_utils::msgs::BAD_CALL, 
                     "get_view_info_from_pname");
        return undef;
    }
    
    unless (file_test(FILE_TEST_EXISTS, $pname)) {
        pretty_print(MSG_ERR, 
                     ClearCase::RD_utils::msgs::PNAME_DOESNT_EXIST,
                     $pname);
        return undef;
    }

    my ($volume, $dirs, $leaf) = splitpath($pname);
    my $deepest_view_dir = undef;
    my $pname_is_dir = file_test(FILE_TEST_DIRECTORY, $pname);

    if ($pname_is_dir) {
        $deepest_view_dir = $pname;
    }
    else {
        $deepest_view_dir = canonpath(catpath($volume, $dirs));
    }

    my $orig_dir = RD_getcwd();

    # Change into the deepest possible directory in the path so we can
    # determine whether we're in a view.
    if (!RD_chdir($deepest_view_dir)) {
        pretty_print(MSG_ERR, 
                     ClearCase::RD_utils::msgs::DIR_ACCESS, 
                     $deepest_view_dir, 
                     "$!");
        $errno = $ERRNO_ACCESS;
        return undef;
    }

    my $view_dir = undef;
    my $error = undef;
    my $view_tag = undef;
    my $dynamic_view = undef;
    my $set_and_wd_match = undef;
    my $view_change_idx = -1;
    my @components = splitdir($dirs);

    # If $pname refers to a directory, then $leaf contains an additional
    # subdirectory that we MUST include as a component.
    if ($leaf && $pname_is_dir) {
        $components[scalar(@components)] = $leaf;
    }

    while (1) {
        # Ensure that we're in a view context.
        $view_tag = get_cwv_tag();

        if (!$view_tag) {
            # Function already blurted if $view_tag is undef.
            if (defined($view_tag)) {
                pretty_print(MSG_ERR, 
                             ClearCase::RD_utils::msgs::PATH_NOT_IN_VIEW,
                             $pname);
                $errno = $ERRNO_NOT_IN_VIEW;
                $error = 1;
            }
            last;
        }

        # Determine whether this view is dynamic.  If it is, we've got its
        # view directory.
        ($dynamic_view, $view_dir) = dynamic_view_is_started($view_tag);
        
        if ($dynamic_view) {
            my $pwv_result = undef;
            
            return undef unless do_cmd("$cleartool pwv -s -setview", 
                                       $pwv_result);
            $set_and_wd_match = ($pwv_result eq $view_tag);
            
            # No point in continuing into the for loop below; we'll never see
            # a change in apparent view
            last if $set_and_wd_match;
        }

        # Let's work our way down from the root directory.  In each successively
        # deeper directory, we'll check if we've entered the view with the
        # matching tag returned from our check in the deepest directory in the
        # input pathname.  
        for ($view_change_idx = 0; 
             $view_change_idx < scalar(@components); 
             $view_change_idx++) 
        {
            my $dir = catdir(@components[0..$view_change_idx]);

            $dir = canonpath(catpath($volume, $dir));

            if (!RD_chdir($dir)) {
                pretty_print(MSG_ERR, 
                             ClearCase::RD_utils::msgs::DIR_ACCESS, 
                             $dir, 
                             "$!");
                $errno = $ERRNO_ACCESS;
                $error = 1;
                last;
            }
                
            my $curr_view_tag = get_cwv_tag();
                
            if ($curr_view_tag eq $view_tag) {
                last;
            }

        } # for each component in the input pathname

        last;

    } # end CWD-preserving while loop

    # Put the CWD back the way we found it.
    if (!RD_chdir($orig_dir)) {
        pretty_print(MSG_WARN, 
                     ClearCase::RD_utils::msgs::DIR_ACCESS, 
                     $orig_dir, 
                     "$!");
    }

    return undef if $error;

    # 1. If it's a snapshot view, we must have entered it at some point in the
    #    loop above, and $view_change_idx points to the last path component
    #    that makes up its directory.
    # 2. If the pathname's working directory view does NOT match the set view
    #    we can use algorithm #1 (because at some point we will have moved
    #    into the working directory context).
    # 3. If the identical working dir & set view tags are on a Windows system,
    #    we must be in a drive mapped to a view.  We can construct the
    #    $view_dir from dynamic_view_is_started(); the $pname_in_view is
    #    everything after the volume/drive letter.
    # 4. On a Unix system, there is no drive letter/volume concept, but
    #    pathnames are case-sensitive, so we may compare the dynamic
    #    view's computed directory with the root component(s) of the input
    #    pathname.  If the view directory is found at the beginning of the
    #    pathname, then everything that follows must be $pname_in_view,
    #    because we have been given a pathname of the "/viewroot/tag/*" form.
    #    Otherwise, the pathname's deepest directory resolved to a view
    #    because this tool is being running within a set view, and so the entire
    #    pathname MUST be view-root-relative.
    while (1) {
        # Case 1 - snapshot view; $view_dir will be populated with contents of
        # @components up to and including $view_change_idx, $pname_in_view
        # will be everything after $view_change_idx
        unless ($dynamic_view) {
            my $dir = catdir(@components[0..$view_change_idx]);
            $view_dir = canonpath(catpath($volume, $dir));
            last;
        }
        
        # Case 2 - dynamic view, but set view != working directory view;
        # $view_dir already contains the correct directory, and $pname_in_view
        # is calculated as in case 1.
        last unless $set_and_wd_match;

        # Case 3 - Windows dynamic view mapped to a drive letter; $view_dir
        # populated as in case 2, and $view_change_idx was initialized to -1
        # and never touched, so $pname_in_view calculated as in case 1 also
        # works.
        last unless os_is_unix();

        # Case 4 - need to determine whether we've been given the set view
        # path or the mvfsroot-rooted path.
        my @view_dir_components = splitdir($view_dir);
        my $i = undef;

        for ($i = 0; 
             ($i < scalar(@view_dir_components)) && ($i < scalar(@components));
             $i++)
        {
            last unless ($components[$i] eq $view_dir_components[$i]);
        }
        
        # If we succesfully matched the entire view directory, then the input
        # pathname contains the view directory.  Otherwise, the entire path
        # makes up the $pname_in_view.
        if ($i == scalar(@view_dir_components)) {
            $view_change_idx = $i - 1;
        }
        else {
            $view_change_idx = -1;
        }

        last;
    }

    my $dir = catdir(@components[($view_change_idx + 1)..$#components]);

    unless (file_name_is_absolute($dir)) {
        $dir = catdir(rootdir(), $dir);
    }
                     
    # Since we already append $leaf onto @components when $pname is a
    # directory, we don't want to include it again...
    my $pname_in_view = canonpath(catpath(undef, 
                                          $dir, 
                                          ($pname_is_dir ? undef : $leaf)));

    return ($view_tag, $view_dir, $pname_in_view);

} # get_view_info_from_pname


################################################################################
# cat_view_dir_and_pname - 
#                         Combine a view directory and pathname into a single
#                         path.
# IN   pname              Candidate pathname.  Assumed to be view-root-relative
#                         (eg. /vobs/src/foo.cxx, \src\foo.cxx).
# IN   view_dir           View directory.
# RESULT:                 Platform-appropriate /view_dir/pname
sub cat_view_dir_and_pname($$) 
{
    my ($pname, $view_dir) = @_;
    return canonpath(catfile($view_dir, $pname));
} # cat_view_dir_and_pname


################################################################################
# chase_symlinks -        Chase a file's symlink chain to produce a pathname 
#                         to an actual file.
#
# IN   pname              Candidate pathname.  Assumed to be view-root-relative
#                         (eg. /vobs/src/foo.cxx, \src\foo.cxx).
# IN   view_dir           View directory in which to attempt to resolve pname.
# RESULT:                 Absolute path that may be interepreted as
#                         view-root-relative (although the symlink text may 
#                         point to anything, anywhere).  May contain ".."
#                         components.  Blurts, returns undef and sets $errno
#                         if errors occur.
sub chase_symlinks($$) 
{
    # Default error case
    $errno = $ERRNO_INT_ERR;

    my ($pname, $view_dir) = @_;
    
    if (!$pname || !$view_dir) {
        pretty_print(MSG_ERR, 
                     ClearCase::RD_utils::msgs::BAD_CALL, 
                     "chase_symlinks");
        return undef;
    }

    # Preliminary check for file existence: if the file is the beginning of
    # a symlink chain and any of the links in the chain are missing,
    # appropriate errors will occur in the loop below; doing a check here
    # allows us to handle the initial file-is-missing case gracefully without
    # incurring the overhead of checking explicitly for file existence at
    # every loop iteration.
    unless (file_test(FILE_TEST_EXISTS, 
                      cat_view_dir_and_pname($pname, $view_dir))) 
    {
        pretty_print(MSG_ERR, 
                     ClearCase::RD_utils::msgs::PNAME_DOESNT_EXIST,
                     $pname);
        return undef;
    }

    my $num_links = 0;
    my $curr_pname = $pname;

    # Chase symlinks, but not forever.  MAXSYMLINKS isn't uniform across our 
    # supported platforms, so choose this number to handle the largest value.
    while ($num_links < 32) {
        my $full_pname = cat_view_dir_and_pname($curr_pname, $view_dir);
        my $link_target = undef;

        if (os_is_unix()) {
            # Use Perl's builtin resolution.  This will work on CC and OS
            # native symlinks.
            return $curr_pname unless (file_test(FILE_TEST_LINK, $full_pname));

            $link_target = readlink($full_pname);
            if (!$link_target) {
                pretty_print(MSG_ERR, 
                             ClearCase::RD_utils::msgs::CANT_CHASE_LINK,
                             $full_pname,
                             "$!");
                return undef;
            }
        }
        else {
            # The builtin Perl -l test doesn't work on Windows because that
            # platform doesn't have the Unix symlink concept, so we rely on
            # cleartool ls output to tell us.  (It would be nicer to use
            # cleartool describe -fmt "%[slink_text]p", but that function
            # interprets symlinks in Windows snapshot views as view-private
            # files, rendering no text for them.)
            #
            # If it's a symlink, the output format is:
            #     pathname --> target [optional commentary]
            #
            # Because square brackets and spaces are acceptable characters in
            # file names, it's almost impossible to consistently determine
            # where the target ends and the optional commentary begins.
            # Luckily, 'pathname --> target' is emitted from a separate
            # message catalog than is ' [optional commentary]', so we can use
            # the message tagging test EV to emit parameterized output.
            #
            # To be polite, we'll save the previous value of the message EV,
            # but it's not strictly necessary; we'd never have gotten this far
            # in this function with message tagging enabled!
            my $old_ev = $ENV{CLEARCASE_ENABLE_ATE_MESSAGES};
            
            $ENV{CLEARCASE_ENABLE_ATE_MESSAGES} = 'True';

            my $ls_output = undef;
            my $cmd = "$cleartool ls \"$full_pname\"";
            my $cmd_success = do_cmd($cmd, $ls_output);
            
            # Replace old EV before any return!
            if ($old_ev) {
                $ENV{CLEARCASE_ENABLE_ATE_MESSAGES} = $old_ev;
            }
            else {
                delete($ENV{CLEARCASE_ENABLE_ATE_MESSAGES});
            }
            
            unless ($cmd_success) {
                return undef;
            }
            
            # Split the result on newlines; the info that we want is always on
            # the 4th line:
            #     TAG_VALUES:pathname --> target:END_VALUES
            my @lines = split(/^/, $ls_output);

            $ls_output = $lines[3];
            $ls_output =~ s/:END_VALUES$//;

            # The pathname and the target can both have embedded whitespace,
            # so the best thing to split on is the arrow.
            my @tokens = split(' --> ', $ls_output);
            my $num_tokens = scalar(@tokens);

            if ($num_tokens == 1) {
                # No arrow ==> no link ==> $curr_pname is the ultimate link
                # target
                return $curr_pname;
            }
            
            $link_target = $tokens[1];
            if (!$link_target || ($num_tokens != 2)) {
                pretty_print(MSG_ERR, 
                             ClearCase::RD_utils::msgs::UNEXPECTED_OUTPUT, 
                             $cmd);
                return undef;
            }

            remove_external_whitespace($link_target);
        }

        $num_links++;        

        # We can go no further if the link target is absolute because we have
        # no way of knowing if the path is outside of the view directory.  
        # ClearCase documentation already suggests using ONLY relative links,
        # so this is the best we can do.
        return $link_target if file_name_is_absolute($link_target);

        my ($volume, $dirs, $file) = splitpath($curr_pname);
        $curr_pname = canonpath(catpath($volume, $dirs, $link_target));
    }

    pretty_print(MSG_ERR, ClearCase::RD_utils::msgs::TOO_MANY_LINKS, $pname);
    $errno = $ERRNO_LOOP;
    return undef;

} # chase_symlinks


# Enumeration for get_object_meta_type output.  Note that VERSION_MT includes 
# regular and derived object versions.
use constant VERSION_MT => 1;
use constant DIR_VERSION_MT => 2;
use constant VIEW_PVT_MT => 3;
use constant NULL_MT => 4;
use constant SOME_OTHER_MT => 5;


################################################################################
# get_object_meta_type -  Get a filesystem object's ClearCase meta-type.
#
# IN   pname              Pathname to desired element version selected by
#                         current view.  Assumed to be in a view context.
# RESULT:                 One of the *_MT values defined above.
#                         Blurts and returns undef if errors occur.
sub get_object_meta_type($) 
{
    my ($pname) = @_;

    if (!$pname) {
        pretty_print(MSG_ERR,
                     ClearCase::RD_utils::msgs::BAD_CALL, 
                     "get_object_meta_type");
        return undef;
    }

    my $cc_obj_kind = undef;
    
    return undef unless 
        do_cmd("$cleartool des -fmt \%m \"$pname\"", $cc_obj_kind);
    
    # "cleartool des -fmt %m" produces output that is never translated, so
    # it's safe to compare explicitly against the string
    if (($cc_obj_kind eq 'derived object version') ||
        ($cc_obj_kind eq 'version'))
    {
        return VERSION_MT;
    }
    elsif ($cc_obj_kind eq 'directory version') {
        return DIR_VERSION_MT;
    }
    elsif ($cc_obj_kind eq 'view private object') {
        return VIEW_PVT_MT;
    }
    elsif ($cc_obj_kind eq '**null meta type**') {
        return NULL_MT;
    }
    else {
        return SOME_OTHER_MT;
    }
} # get_object_meta_type


################################################################################
# convert_pname_to_oref - Convert a version pathname to a OREF.  Assumes 
#                         that caller has ensured a view context and that the
#                         input pathname refers to a checked-in ClearCase FILE
#                         version.
#
# IN   pname              Pathname to desired element version selected by
#                         current view.  Assumed to be view-root-relative
#                         (eg. /vobs/src/foo.cxx, \src\foo.cxx).
# IN   view_dir           View directory in which to attempt to resolve pname.
# RESULT:                 (OREF string, config spec rule) in list context.
#                         OREF string only in scalar context.
#                         Blurts, returns undef and sets $errno if errors
#                         occur. 
sub convert_pname_to_oref($$)
{
    # Default error case
    $errno = $ERRNO_INT_ERR;

    my ($pname, $view_dir) = @_;

    if (!$pname || !$view_dir) {
        pretty_print(MSG_ERR, 
                     ClearCase::RD_utils::msgs::BAD_CALL, 
                     "convert_pname_to_oref");
        return undef;
    }

    # Need to chase soft links; they don't register as vob objects, so we
    # can't ensure that they're checked in.  (Hard links are okay in this
    # regard.)  
    my $real_pname = chase_symlinks($pname, $view_dir);
    if (!$real_pname) {
        return undef;
    }

    $real_pname = cat_view_dir_and_pname($real_pname, $view_dir);

    # Ensure that item is not a directory
    if (file_test(FILE_TEST_DIRECTORY, $real_pname)) {
        pretty_print(MSG_ERR, 
                     ClearCase::RD_utils::msgs::DIR_NOT_ALLOWED, 
                     $pname);
        $errno = $ERRNO_IS_DIRECTORY;
        return undef;
    }

    # Ensure that item is a known meta-type.
    my $meta_type = get_object_meta_type($real_pname);
    
    return undef unless $meta_type;

    # View-private files get special output, as we expect them to be the
    # most frequent occurrence.
    if ($meta_type == VIEW_PVT_MT) {
        pretty_print(MSG_ERR,
                     ClearCase::RD_utils::msgs::VIEW_PVT_NOT_ALLOWED,
                     $pname);
        $errno = $ERRNO_VIEW_PVT;
        return undef;
    }
    elsif ($meta_type != VERSION_MT) {
        pretty_print(MSG_ERR,
                     ClearCase::RD_utils::msgs::NON_VERSION_NOT_ALLOWED,
                     $pname);
        $errno = $ERRNO_NOT_VERSION_FILE;
        return undef;
    }
    
    # Ensure that item is not checked out.
    my $checkout_spec = undef;

    if (!do_cmd("$cleartool lsco -cview \"$real_pname\"", $checkout_spec)) {
        return undef;
    }
    if ($checkout_spec) {
        pretty_print(MSG_ERR, ClearCase::RD_utils::msgs::NO_CHECKOUTS, $pname);
        $errno = $ERRNO_CHECKOUT;
        return undef;
    }

    # retrieve OREF (only output should be the reference)
    my $oref = undef;

    return undef unless 
        do_cmd("$cleartool des -fmt \"%[universal_selector]p\" \"$real_pname\"",
               $oref);

    return $oref unless wantarray();

    # Get config spec rule 
    my $ls_output = undef;
    my $cmd = "$cleartool ls -nxname \"$real_pname\"";

    return undef if (!do_cmd($cmd, $ls_output));

    # CS rule is the last item in the line after whitespace:
    #     /vobs/src/foo.cxx        Rule: /main/LATEST
    # The rule consists of branch names separated by / or \ (depending on
    # platform), but it CANNOT contain the ": " that separates the (possibly
    # translated) "Rule" string from the rule itself.
    my @ls_tokens = split(/: /, $ls_output);
    my $rule = $ls_tokens[$#ls_tokens];

    if (!$rule) {
        pretty_print(MSG_ERR, 
                     ClearCase::RD_utils::msgs::UNEXPECTED_OUTPUT, 
                     $cmd);
        return undef;
    }
    
    # Eliminate leading/trailing whitespace & newlines
    chomp($rule);
    remove_external_whitespace($rule);

    return ($oref, $rule);

} # convert_pname_to_oref


################################################################################
# convert_oref_to_pname - Convert a OREF to a version pathname.  Assumes 
#                         that caller has ensured a view context.
#
# IN   oref               Obejct's OREF
# RESULT:                 View-extended pathname of object (blurts and returns 
#                         undef if errors occur)
sub convert_oref_to_pname($)
{
    my ($oref) = @_;

    if (!$oref) {
        pretty_print(MSG_ERR, 
                     ClearCase::RD_utils::msgs::BAD_CALL, 
                     "convert_oref_to_pname");
        return undef;
    }

    my $pname = undef;

    if (!do_cmd("$cleartool des -s $oref", $pname)) {
        return undef;
    }
 
    return $pname;

} # convert_oref_to_pname

# OREF URI data
use constant URI_SCHEME => "ccase";
use constant OREF_QUERY_NAME => "OREF";
use constant RGY_QUERY_NAME => "registry_server";
use constant REGION_QUERY_NAME => "region";

################################################################################
# convert_pname_to_uri -  Convert a pathname to a URI string.
#          
# IN   pname              If view_dir is specified, a view-root-relative 
#                         pathname (eg. /vobs/src/foo.cxx) to a ClearCase FILE
#                         version.
#                         If view_dir is NOT specified, an absolute pathname
#                         (eg. /view/my_view/vobs/src/foo.cxx) to a ClearCase
#                         FILE version.
# IN   view_dir           Optional directory that specifies the view to be
#                         used to generate the URI 
#                         (eg. /<mvfsroot>/<dynamic view tag> or 
#                         <snapshot view directory>).
# RESULT:                 Internal-UTF-8 (URI string, config spec rule) in
#                         list context; URI string only in scalar context.
#                         Note that the URI string, although encoded in
#                         internal UTF-8, will ALWAYS consist entirely of
#                         ASCII characters.  It is returned in UTF-8 only to
#                         be consistent with the config spec rule, which may 
#                         contain non-ASCII characters.
#                         Blurts, returns undef and sets $errno if errors
#                         occur. 
sub convert_pname_to_uri($$)
{
    # Default error case
    $errno = $ERRNO_INT_ERR;

    my ($pname, $view_dir) = @_;

    if (!$pname) {
        pretty_print(MSG_ERR, 
                     ClearCase::RD_utils::msgs::BAD_CALL,
                     "convert_pname_to_uri");
        return undef;
    }

    if (!$view_dir) {
        (my $view_tag, $view_dir, $pname) = get_view_info_from_pname($pname);
        return undef if (!$view_tag);
    }

    my ($native_oref, $config_spec_rule) = 
        convert_pname_to_oref($pname, $view_dir);
    
    if (!$native_oref || !$config_spec_rule) {
        return undef;
    }

    my $uri_obj = new URI();
    my ($registry, $region) = get_registry_and_region();

    unless ($registry && $region) {
        return undef;
    }

    $uri_obj->scheme(URI_SCHEME);

    # $pname may have '\' characters that we want to transpose to '/' for use
    # in the URI.
    my $utf8_pname = native_cp_to_internal_utf8($pname);

    return undef unless ($utf8_pname);
    $utf8_pname =~ s/\\/\//g;

    # Need to encode $pname into external UTF-8 so that it will treat contents
    # as bytes and percent-escape them accordingly.
    eval { $utf8_pname = encode(UTF8_ENCODING, $utf8_pname, Encode::FB_CROAK) };
    if ($@) {
        pretty_print(MSG_ERR,
                     ClearCase::RD_utils::msgs::CANT_CONVERT_FROM_UTF8,
                     $pname,
                     UTF8_ENCODING,
                     $@);
        return undef;
    }
    
    $uri_obj->path($utf8_pname);
    $uri_obj->query_form(OREF_QUERY_NAME() => $native_oref,
                         RGY_QUERY_NAME() => $registry,
                         REGION_QUERY_NAME() => $region);

    my $uri_str = 
        native_cp_to_internal_utf8($uri_obj->canonical()->as_string());
    
    return (wantarray() ? 
            ($uri_str, native_cp_to_internal_utf8($config_spec_rule)) : 
            $uri_str);

} # convert_pname_to_uri


################################################################################
# get_oref_from_uri -     Get the OREF string from a URI.
#          
# IN   uri_str            URI string.
# RESULT:                 OREF string. Blurts and returns undef if errors
#                         occur.
sub get_oref_from_uri($)
{
    my ($uri_str) = @_;

    if (!$uri_str) {
        pretty_print(MSG_ERR, 
                     ClearCase::RD_utils::msgs::BAD_CALL,
                     "get_oref_from_uri");
        return undef;
    }

    # Don't need to decode anything here; when we created the URI in 
    # convert_pname_to_uri, we encoded only the authority & path components,
    # and we don't need them here.

    my $uri_obj = new URI($uri_str);
    my $actual_scheme = $uri_obj->scheme();

    if ($actual_scheme ne URI_SCHEME) {
        pretty_print(MSG_ERR, 
                     ClearCase::RD_utils::msgs::BAD_URI_SCHEME, 
                     $actual_scheme);
        return undef;
    }

    # Report ALL problems found with the URI before returning.
    my $num_errors = 0;
    my %query_hash = $uri_obj->query_form();

    # Check for existence of required query NV pairs.
    foreach my $name (OREF_QUERY_NAME) {
        next if ($query_hash{$name});
        pretty_print(MSG_ERR, 
                     ClearCase::RD_utils::msgs::MISSING_URI_QUERY_NAME,
                     $name);
        $num_errors++;
    }

    return ($num_errors > 0 ? undef : $query_hash{OREF_QUERY_NAME()});
} # get_oref_from_uri


################################################################################
# convert_uri_to_pname -  Convert a URI string to a version-extended pathname.
#          
# IN   uri_str            URI string.
# RESULT:                 Version-extended pathname from the current view
#                         context (blurts and returns undef if errors occur).
sub convert_uri_to_pname($)
{
    my ($uri_str) = @_;
    my $oref_str = get_oref_from_uri($uri_str);

    return ($oref_str ? convert_oref_to_pname($oref_str) : undef);
} # convert_uri_to_pname


################################################################################
# get_oref_info_from_uri -
#                         Get the version OID and vob UUID from the OREF
#                         embedded in a URI.
#          
# IN   uri_str            URI string.
# RESULT:                 Array containing ($version_oid, $vob_uuid).
#                         Blurts and returns undef if errors occur.
sub get_oref_info_from_uri($)
{
    my ($uri_str) = @_;
    my $oref_str = get_oref_from_uri($uri_str);

    return undef unless $oref_str;
    
    # ClearCase produces "oid:<version-oid>@vobuuid:<vob-uuid>".  We know that
    # it doesn't contain any other instances of "oid:" or "@vobuuid:".
    
    my @tokens = split(/\@vobuuid:/, $oref_str);

    if (scalar(@tokens) == 2) {
        my $version_oid = $tokens[0];
        
        $version_oid =~ s/^oid://;
        return ($version_oid, $tokens[1]);
    }

    pretty_print(MSG_ERR, 
                 ClearCase::RD_utils::msgs::MALFORMED_URI_OREF, 
                 $uri_str);
    return undef;
    
} # get_oref_info_from_uri


my $windows_os = undef;

################################################################################
# os_is_unix -            Is this a Unix OS?  (Implies support for symlinks,
#                         case-sensitive pathnames and set views.)
#
# RESULT:                 True iff this is a Unix OS.
sub os_is_unix()
{
    if (!defined($windows_os)) {
        my $arch = $^O ? $^O : $Config{archname};
        $windows_os = (($arch =~ /^MSWin32/) || 
                       ($arch eq 'i386-win32')) ? 
                       1 : 0;
    }
    return !$windows_os;
} # os_is_unix


my @baselines = ();

################################################################################
# append_baseline_to_buffer - 
#                         Helper function for get_baselines_from_du_file.
#                         Appends baselines parsed by that function to
#                         @baselines, which that function is responsible for 
#                         initializing.
#
# IN   expat              Underlying expat instance from XML::Parser.
# IN   eltype             Element type (the tag's name).
sub append_baseline_to_buffer($$%) 
{
    my ($expat, $eltype, %attrs) = @_;
    my $baseline_text = $attrs{BASELINE_NAME_ATTR()};

    return unless (($eltype eq XML_BASELINE_ELEMENT) && 
                   defined($baseline_text));

    $baselines[scalar(@baselines)] = $baseline_text;
} # append_baseline_to_buffer


################################################################################
# get_baselines_from_du_file -
#                         Parse a deployment unit file to retrieve its
#                         baselines.
#
# IN   du_file            Path to deployment unit file.
# RESULT:                 Potentially empty array of baseline strings in
#                         internal UTF-8.  Blurts, sets errno and dies if
#                         errors occur.
sub get_baselines_from_du_file($)
{
    # Default error case
    $errno = $ERRNO_INT_ERR;

    my ($du_file) = @_;
    
    unless ($du_file) {
        pretty_print(MSG_ERR, 
                     ClearCase::RD_utils::msgs::BAD_CALL,
                     "get_baselines_from_du_file");
        die;
    }

    @baselines = ();

    my $parser = new XML::Parser(Handlers => 
                                 {Start => \&append_baseline_to_buffer});
    
    eval {
        $parser->parsefile($du_file);
    };
    if ($@) {
        pretty_print(MSG_ERR, ClearCase::RD_utils::msgs::BAD_FILE, "$@");
        die;
    }
 
    return @baselines;

} # get_baselines_from_du_file


################################################################################
# internal_utf8_to_native_cp -
#                         Convert a string from Perl's internal UTF-8
#                         representation to the system's native codepage.
#
# IN   str                String to convert.
# RESULT:                 Natively-encoded string.  Blurts and returns undef
#                         if errors occur.
sub internal_utf8_to_native_cp($)
{
    my ($str) = @_;

    $str = eval { encode(get_codepage(), $str, Encode::FB_CROAK) };
    return $str unless ($@);

    pretty_print(MSG_ERR, 
                 ClearCase::RD_utils::msgs::CANT_CONVERT_FROM_UTF8,
                 $str,
                 get_codepage(),
                 $@);
    return undef;

} # internal_utf8_to_native_cp


################################################################################
# native_cp_to_internal_utf8 -
#                         Convert a string from the system's native codepage
#                         to Perl's internal UTF-8 representation.
#
# IN   str                String to convert.
# RESULT:                 Internal-UTF-8-encoded string.  Blurts and returns
#                         undef if errors occur.
sub native_cp_to_internal_utf8($)
{
    my ($str) = @_;

    $str = eval { decode(get_codepage(), $str, Encode::FB_CROAK) };
    return $str unless ($@);

    pretty_print(MSG_ERR, 
                 ClearCase::RD_utils::msgs::CANT_CONVERT_TO_UTF8,
                 $str,
                 get_codepage(),
                 $@);
    return undef;
} # native_cp_to_internal_utf8


my $codepage = undef;

################################################################################
# get_codepage -          Get this system's native codepage.
#
# RESULT:                 An encoding usable by the Encode package, binmode,
#                         etc.  Dies if unable to determine the codepage.
sub get_codepage() 
{
    return $codepage if $codepage;

    if (os_is_unix()) {
        # Need to examine the locale setting.
        use POSIX;  
        import POSIX qw(LC_CTYPE);

        # Don't panic!  With no second argument, setlocale() queries.
        for (POSIX::setlocale(LC_CTYPE)) {
            chomp;

            # ANY locale string with "UTF-8" means that character data is
            # encoded in UTF-8.
            /UTF-8/ &&
                do { $codepage = 'UTF-8'; last; };

            (/^zh_HK/ || /^zh_TW/) &&
                do { $codepage = 'big5'; last; };

            /^ja/ &&
                do { $codepage = 'Shift_JIS'; last; };

            /^ko/ &&
                do { $codepage = 'EUC-KR'; last; };

            (/^zh_CN/ || /^zh_SG/) &&
                do { $codepage = 'GBK'; last; };
            
            /^vi/ &&
                do { $codepage = 'cp1258'; last; };
            
            /^th/ &&
                do { $codepage = 'cp874'; last; };
                
            /^et/ && 
                do { $codepage = 'cp1257'; last; };

            # This is the only en_ locale that doesn't use 8859-1.
            (/^ar/ || /^en_PK/) && 
                do { $codepage = 'ISO-8859-6'; last; };
            
            # Don't put this BEFORE 8859-6 above, or else the /^en/ test will
            # gobble up en_PK.
            (/^da/ || /^de/ || /^en/ || /^es/ || /^fi/ || 
             /^fr/ || /^id/ || /^is/ || /^it/ || /^ms/ || 
             /^nb/ || /^nl/ || /^pt/ || /^sq/ || /^sv/) &&
                do { $codepage = 'ISO-8859-1'; last; };

            (/^cs/ || /^hr/ || /^hu/ || /^pl/ || 
             /^ro/ || /^sh/ || /^sk/ || /^sl/) && 
                do { $codepage = 'ISO-8859-2'; last; };

            (/^be/ || /^bg/ || /^mk/ || /^ru/ || /^sr/ || /^uk/) && 
                do { $codepage = 'ISO-8859-5'; last; };

            /^el/ && 
                do { $codepage = 'ISO-8859-7'; last; };

            /^he/ &&
                do { $codepage = 'ISO-8859-8'; last; };

            /^tr/ &&
                do { $codepage = 'ISO-8859-9'; last; };

            (/^lt/ || /^lv/) && 
                do { $codepage = 'ISO-8859-13'; last; };

            # Default locale (ie. 'C' or anything not explicitly called out
            # above) is ASCII.
            $codepage = 'ASCII';
        }
    }
    else {
        # System codepage lives in the Windows registry.
        require Win32::TieRegistry;
        
        my $regkey = Win32::TieRegistry->new('HKEY_LOCAL_MACHINE/SYSTEM/' .
                                             'CurrentControlSet/Control/' .
                                             'Nls/CodePage',
                                             {Delimiter => '/'});

        # Registry value is a number (ie. 1252); to be used by the Encode
        # package, it must include 'cp' (cp1252).
        $codepage = 'cp' . 
            lc($regkey->GetValue('ACP') || $regkey->GetValue('OEMCP'));
    }
    
    return $codepage;
    
} # get_codepage


################################################################################
# scalar_in_scalar_out -  Generic UTF-8 function for use by
#                         File::Spec::Functions wrappers.
#
# IN   function_ref       The function to be called.
# IN   native_str         The string to be converted to UTF-8 and then passed
#                         to function_ref.
# IN   pass_thru_args     Additional arguments to be passed through to
#                         function_ref.
# RESULT:                 The scalar result of $function_ref in the native
#                         codepage.
sub scalar_in_scalar_out($$@)
{
    my ($function_ref, $native_str, @pass_thru_args) = @_;
    my $utf8_str = native_cp_to_internal_utf8($native_str);
    return 
        internal_utf8_to_native_cp(&$function_ref($utf8_str, @pass_thru_args));
} # scalar_in_scalar_out


################################################################################
# array_in_scalar_out -   Generic UTF-8 function for use by
#                         File::Spec::Functions wrappers.
#
# IN   function_ref       The function to be called.
# IN   native_strs        The strings to be converted to UTF-8 and then passed
#                         to function_ref.
# RESULT:                 The scalar result of $function_ref in the native
#                         codepage.
sub array_in_scalar_out($@)
{
    my ($function_ref, @native_strs) = @_;
    my @utf8_strs = ();
    foreach my $str (@native_strs) {
        push(@utf8_strs, native_cp_to_internal_utf8($str));
    }
    return internal_utf8_to_native_cp(&$function_ref(@utf8_strs));
} # array_in_scalar_out


################################################################################
# scalar_in_array_out -   Generic UTF-8 function for use by
#                         File::Spec::Functions wrappers.
#
# IN   function_ref       The function to be called.
# IN   native_str         The string to be converted to UTF-8 and then passed
#                         to function_ref.
# IN   pass_thru_args     Additional arguments to be passed through to
#                         function_ref.
# RESULT:                 The array results of $function_ref in the native
#                         codepage.
sub scalar_in_array_out($$@)
{
    my ($function_ref, $native_str, @pass_thru_args) = @_;
    my $utf8_str = native_cp_to_internal_utf8($native_str);
    my @results = &$function_ref($utf8_str, @pass_thru_args);
    foreach my $str (@results) {
        $str = internal_utf8_to_native_cp($str);
    }
    return @results;
} # scalar_in_array_out


################################################################################
# File::Spec::Functions wrappers
################################################################################
#
# File::Spec::Functions functions don't play very nicely with multibyte
# character sets on Windows because they cannot distinguish between embedded
# 0x5c bytes and the single-byte '\' file separator character.  The general
# solution to this problem is to wrap calls to all such functions so that the
# inputs are first converted to UTF-8.
#
# Usage for these functions can be found in the File::Spec::Functions
# documentation.
#
use File::Spec::Functions ();

sub canonpath 
{
    my $fn = \&File::Spec::Functions::canonpath;
    my $scalar_in = shift(@_);
    return os_is_unix() ? 
        &$fn($scalar_in, @_) : scalar_in_scalar_out($fn, $scalar_in, @_);
} # canonpath

sub catdir
{
    my $fn = \&File::Spec::Functions::catdir;
    os_is_unix() && return &$fn(@_);
    my $result = array_in_scalar_out($fn, @_);
    if (defined($result) && !$result) {
        $result = rootdir();
    }
    return $result;
} # catdir

sub catfile
{
    my $fn = \&File::Spec::Functions::catfile;
    return os_is_unix() ? &$fn(@_) : array_in_scalar_out($fn, @_);
} # catfile

sub catpath
{
    my $fn = \&File::Spec::Functions::catpath;
    return os_is_unix() ? &$fn(@_) : array_in_scalar_out($fn, @_);
} # catpath

sub file_name_is_absolute
{
    my $file_name = shift @_;
    unless (os_is_unix()) {
        $file_name = native_cp_to_internal_utf8($file_name);
    }
    return File::Spec::Functions::file_name_is_absolute($file_name, @_);
} # file_name_is_absolute

sub rel2abs
{
    my ($path, $base) = @_;
    
    os_is_unix() && return File::Spec::Functions::rel2abs($path, $base);

    if ($base) {
        $base = rel2abs($base) unless (file_name_is_absolute($base));
    }
    else {
        $base = RD_getcwd();
    }
    return array_in_scalar_out(\&File::Spec::Functions::rel2abs, $path, $base);
} # rel2abs

sub rootdir
{
    return File::Spec::Functions::rootdir(@_);
} # rootdir

sub splitdir
{
    my $fn = \&File::Spec::Functions::splitdir;
    my $scalar_in = shift(@_);
    return os_is_unix() ? 
        &$fn($scalar_in, @_) : scalar_in_array_out($fn, $scalar_in, @_);
} # splitdir

sub splitpath
{
    my $fn = \&File::Spec::Functions::splitpath;
    my $scalar_in = shift(@_);
    return os_is_unix() ? 
        &$fn($scalar_in, @_) : scalar_in_array_out($fn, $scalar_in, @_);
} # splitpath


################################################################################
# Cwd wrappers
################################################################################
#
# Cwd functions don't play very nicely with multibyte character sets on
# Windows because they cannot distinguish between embedded 0x5c bytes and the
# single-byte '\' file separator character.  There is no "general" solution.
#
# Usage for these functions can be found in the Cwd documentation.
#
use Cwd ();

sub RD_chdir
{
    # Unix works as-is.
    os_is_unix() && return Cwd::chdir(@_);
    
    my $tgt_dir = shift(@_);

    chomp($tgt_dir);
    remove_external_whitespace($tgt_dir);
    
    # We need to escape any trailing 0x5c byte.  Embedded bytes don't seem to
    # cause any problems.
    $tgt_dir =~ s/\\$/\\\\/;

    return chdir($tgt_dir);
} # RD_chdir

sub RD_getcwd
{
    # Unix works as-is.
    os_is_unix() && return Cwd::getcwd();
    
    # On Windows, the "cd" command with no arguments returns the current
    # working directory.  It works correctly when the CWD contains embedded
    # 0x5c characters, but is known to fail silently when the final byte
    # occurs in the final position.
    #
    # Can't use do_cmd because we don't want any blurting.
    my $cwd = `cd`;

    if ($?) {
        # Cwd::getcwd returns the correct directory, but it gobbles the
        # trailing 0x5c.
        $cwd = Cwd::getcwd() . "\\";
    }
    chomp($cwd);
    return $cwd;
} # RD_getcwd


################################################################################
# unary file test wrappers
################################################################################
#
# The unary file test operators don't play very nicely with multibyte
# character sets on Windows because they cannot distinguish between embedded
# 0x5c bytes and the single-byte '\' file separator character.  There is no
# "general" solution, so not all operators are supported.  
#
#

################################################################################
# file_test -             Wrapper for Perl's builtin unary file test
#                         operators.
#
# IN   test               Test to perform.  FILE_TEST_* enumerated values
#                         represent the ONLY supported operators!
# IN   path               Path to test.
# RESULT:                 True if test passes.  False otherwise.
sub file_test($$) 
{
    my ($test, $path) = @_;

    $path || return undef;

    if (os_is_unix()) {
        if ($test == FILE_TEST_DIRECTORY) {
            return -d $path;
        }
        elsif ($test == FILE_TEST_EXISTS) {
            return -e $path;
        }
        elsif ($test == FILE_TEST_LINK) {
            return -l $path;
        }
        elsif ($test == FILE_TEST_FILE) {
            return -f $path;
        }
    }
    else {
        require Win32::File;
        import Win32::File qw(GetAttributes);

        my $attrs = 0;
        
        # If the path doesn't exist, this function fails.
        return undef unless GetAttributes($path, $attrs);
        
        my $is_dir = ($attrs & Win32::File::DIRECTORY()) ? 1 : undef;
        
        if ($test == FILE_TEST_DIRECTORY) {
            return $is_dir;
        }
        elsif ($test == FILE_TEST_EXISTS) {
            return 1; # $path MUST exist if GetAttributes succeeded
        }
        elsif ($test == FILE_TEST_LINK) {
            return undef; # no links on Windows
        }
        elsif ($test == FILE_TEST_FILE) {
            return !$is_dir;
        }
    }

    pretty_print(MSG_ERR, ClearCase::RD_utils::msgs::BAD_CALL, "file_test");
    return undef;

} # file_test

1;
